/* Copyright (c) 2013 tonomu
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>
#include <string.h>
#include "midiminish.h"

static int usgep(int argc);
static int fileo(char *argv[], FILE **ifp, FILE **ofp);
static int midix(FILE *ifp, FILE *ofp);
static int headx(FILE *ifp, FILE *ofp);
static int charx(int count, FILE *ifp, FILE *ofp, unsigned int *value);
static int trckx(FILE *ifp, FILE *ofp);
static int vldtx(FILE *ifp, FILE *ofp, struct track_st *track);
static int evntx(FILE *ifp, FILE *ofp, struct track_st *track);
static int typex(FILE *ifp, FILE *ofp, struct track_st *track);
static int contx(FILE *ifp, FILE *ofp, struct track_st *track);
static int fxmgx(FILE *ifp, FILE *ofp, struct track_st *track);
static int chmgx(FILE *ifp, FILE *ofp, struct track_st *track);
static int datax(int count, FILE *ifp, FILE *ofp, struct track_st *track);
static int sizes(FILE *ofp, unsigned int value);
static int filec(FILE **ifp, FILE **ofp);

int main(int argc, char *argv[])
{
  FILE *ifp;
  FILE *ofp;
  int status;

  status = usgep(argc);
  if (status < 0) { return status; }

  status = fileo(argv, &ifp, &ofp);
  if (status < 0) { return status; }

  status = midix(ifp, ofp);
  if (status < 0) {
    (void) filec(&ifp, &ofp);
    return status;
  }

  status = filec(&ifp, &ofp);
  if (status < 0) { return status; }

  return STATUS_SUCCESS;
}

static int usgep(int argc)
{
  if (argc != 3) {
    (void) printf("midiminish version %d.%d.%d\n",
                   MAJOR_VERSION, MINOR_VERSION, MODIFY_VERSION);
    (void) printf("usage: midiminish SOURCE DEST\n");
    return E_ARGUMENT_ENTRY;
  }

  return STATUS_SUCCESS;
}

static int fileo(char *argv[], FILE **ifp, FILE **ofp)
{
  int status;

  argv++;

  status = strcmp(*argv, *(argv + 1));
  if (status == 0) {
    (void) printf("error: illegal output file\n");
    return E_FILE_ENTRY;
  }

  *ifp = fopen(*argv, "rb");
  if (*ifp == NULL) {
    (void) printf("error: failed to open input file\n");
    return E_OPEN_INPUT;
  }

  argv++;
  *ofp = fopen(*argv, "wb");
  if (*ofp == NULL) {
    (void) printf("error: failed to open output file\n");
    return E_OPEN_OUTPUT;
  }

  return STATUS_SUCCESS;
}

static int midix(FILE *ifp, FILE *ofp)
{
  int total;
  int num;
  int status;

  total = headx(ifp, ofp);
  if (total < 0) {
    (void) printf("error: failed to output header\n");
    return total;
  }

  for (num = 0; num < total; num++) {
    status = trckx(ifp, ofp);
    if (status < 0) {
      (void) printf("error: failed to output track %d\n", num);
      return status;
    }
  }

  return STATUS_SUCCESS;
}

static int headx(FILE *ifp, FILE *ofp)
{
  int status;
  unsigned int total;
  unsigned int value;
  const unsigned int mthd = 0x4d546864U;

  status = charx(4, ifp, ofp, &value);
  if (status < 0) {
    (void) printf("error: failed to output MThd\n");
    return status;
  }
  if (value != mthd) {
    (void) printf("error: unsupported file\n");
    return E_MTHD_VALUE;
  }

  status = charx(4, ifp, ofp, &value);
  if (status < 0) {
    (void) printf("error: failed to output header size\n");
    return status;
  }
  if (value != 6) {
    (void) printf("error: unsupported header size\n");
    return E_HSIZE_VALUE;
  }

  status = charx(2, ifp, ofp, &value);
  if (status < 0) {
    (void) printf("error: failed to output SMF format\n");
    return status;
  }
  if (value >> 1) {
    (void) printf("error: unsupported format\n");
    return E_FORMAT_VALUE;
  }

  status = charx(2, ifp, ofp, &total);
  if (status < 0) {
    (void) printf("error: failed to output track count\n");
    return status;
  }

  status = charx(2, ifp, ofp, &value);
  if (status < 0) {
    (void) printf("error: failed to output TPQN\n");
    return status;
  }

  return (int) total;
}

static int charx(int count, FILE *ifp, FILE *ofp, unsigned int *value)
{
  int status;

  *value = 0;

  for ( ; count; count--) {
    status = fgetc(ifp);
    if (status == EOF) {
      (void) printf("error: failed to fgetc\n");
      return E_GET_CHAR;
    }

    status = fputc(status, ofp);
    if (status == EOF) {
      (void) printf("error: failed to fputc\n");
      return E_PUT_CHAR;
    }

    *value = (*value << 8) | (unsigned int) status;
  }

  return STATUS_SUCCESS;
}

static int trckx(FILE *ifp, FILE *ofp)
{
  int status;
  unsigned int value;
  unsigned int count;
  struct track_st track;
  fpos_t start;
  fpos_t end;
  const unsigned int mtrk = 0x4d54726bU;

  status = charx(4, ifp, ofp, &value);
  if (status < 0) {
    (void) printf("error: failed to output MTrk\n");
    return status;
  }
  if (value != mtrk) {
    (void) printf("error: unsupported track format\n");
    return E_MTRK_VALUE;
  }

  status = fgetpos(ofp, &start);
  if (status != 0) {
    (void) printf("error: failed to get initial data position\n");
    return E_GET_INIDATA;
  }

  status = charx(4, ifp, ofp, &count);
  if (status < 0) {
    (void) printf("error: failed to get track size\n");
    return status;
  }

  track.buf = 0xF0;
  track.in = 0;
  track.out = 0;
  while (track.in < count) {
    status = vldtx(ifp, ofp, &track);
    if (status < 0) {
      (void) printf("error: failed to output time\n");
      return status;
    }

    status = evntx(ifp, ofp, &track);
    if (status < 0) {
      (void) printf("error: failed to output event\n");
      return status;
    }
  }
  if (track.in != count) {
    (void) printf("error: invalid track size\n");
    return E_TRACK_SIZE;
  }
  if (track.in < track.out) {
    (void) printf("error: invalid track data\n");
    return E_TRACK_DATA;
  }

  status = fgetpos(ofp, &end);
  if (status != 0) {
    (void) printf("error: failed to get last data position\n");
    return E_GET_LASTDATA;
  }

  status = fsetpos(ofp, &start);
  if (status != 0) {
    (void) printf("error: failed to set initial data position\n");
    return E_SET_INIDATA;
  }

  status = sizes(ofp, track.out);
  if (status < 0) {
    (void) printf("error: failed to set track size\n");
    return status;
  }

  status = fsetpos(ofp, &end);
  if (status != 0) {
    (void) printf("error: failed to set last data position\n");
    return E_SET_LASTDATA;
  }

  return STATUS_SUCCESS;
}

static int sizes(FILE *ofp, unsigned int value)
{
  int status;
  int i;

  i = 24;
  while (i >= 0) {
    status = fputc((int) ((value >> i) & 0xFF), ofp);
    if (status == EOF) {
      (void) printf("error: failed to fputc\n");
      return E_PUT_CHAR;
    }
    i = i - 8;
  }

  return STATUS_SUCCESS;
}

static int vldtx(FILE *ifp, FILE *ofp, struct track_st *track)
{
  int value;
  int i;
  int status;

  value = 0;

  for (i = 4; i ; i--)
  {
    status = fgetc(ifp);
    if (status == EOF) {
      (void) printf("error: fgetc failure\n");
      return E_GET_CHAR;
    }
    (track->in)++;
    status = fputc(status, ofp);
    if (status == EOF) {
      (void) printf("error: fputc failure\n");
      return E_PUT_CHAR;
    }
    (track->out)++;

    value = value << 7;
    value = value | (status & 0x7f);

    if ((status >> 7) == 0) { return value; }
  }

  (void) printf("error: illegal length format\n");
  return E_LEN_FORMAT;
}

static int evntx(FILE *ifp, FILE *ofp, struct track_st *track)
{
  int status;

  status = typex(ifp, ofp, track);
  if (status < 0) {
    (void) printf("error: failed to output MIDI event\n");
    return status;
  }

  status = contx(ifp, ofp, track);
  if (status < 0) {
    (void) printf("error: failed to output MIDI message\n");
    return status;
  }

  return STATUS_SUCCESS;
}

static int typex(FILE *ifp, FILE *ofp, struct track_st *track)
{
  int type;
  int status;
  int put;

  type = fgetc(ifp);
  if (type == EOF) {
    (void) printf("error: fgetc failure\n");
    return E_GET_CHAR;
  }
  (track->in)++;

  if ((type & 0xF0) == 0xF0) {
    track->buf = type;
    put = 1;
  }
  else if (type < 0x80) {
    status = ungetc(type, ifp);
    if (status == EOF) {
      (void) printf("error: ungetc failure\n");
      return E_UNGET_CHAR;
    }
    (track->in)--;
    if ((track->buf & 0xF0) == 0xF0) {
      (void) printf("error: invalid running status\n");
      return E_RUNNING_STATUS;
    }
    put = 0;
  }
  else if (type != track->buf) {
    track->buf = type;
    put = 1;
  }
  else {
    put = 0;
  }

  if (put) {
    status = fputc(track->buf, ofp);
    if (status == EOF) {
      (void) printf("error: fputc failure\n");
      return E_PUT_CHAR;
    }
    (track->out)++;
  }

  return STATUS_SUCCESS;
}

static int contx(FILE *ifp, FILE *ofp, struct track_st *track)
{
  int status;

  if ((track->buf & 0xF0) == 0xF0) {
    status = fxmgx(ifp, ofp, track);
    if (status < 0) {
      (void) printf("error: failed to output sysex or meta message\n");
      return status;
    }
  }
  else {
    status = chmgx(ifp, ofp, track);
    if (status < 0) {
      (void) printf("error: failed to output channel message\n");
      return status;
    }
  }

  return STATUS_SUCCESS;
}

static int fxmgx(FILE *ifp, FILE *ofp, struct track_st *track)
{
  int count;
  int status;

  if (track->buf == 0xFF) {
    status = datax(1, ifp, ofp, track);
    if (status < 0) {
      (void) printf("error: failed to output meta message type\n");
      return status;
    }
  }
  else if ((track->buf != 0xF0) && (track->buf != 0xF7)) {
    (void) printf("error: unsupported MIDI event\n");
    return E_INS_TYPE;
  }


  count = vldtx(ifp, ofp, track);
  if (count < 0) {
    (void) printf("error: failed to output message length\n");
    return count;
  }

  status = datax(count, ifp, ofp, track);
  if (status < 0) {
    (void) printf("error: failed to output variable length data\n");
    return status;
  }

  return STATUS_SUCCESS;
}

static int chmgx(FILE *ifp, FILE *ofp, struct track_st *track)
{
  int count;
  int status;

  if ((track->buf & 0xE0) == 0xC0) { count = 1; }
  else { count = 2; }

  status = datax(count, ifp, ofp, track);
  if (status < 0) {
    (void) printf("error: failed to output MIDI message\n");
    return status;
  }

  return STATUS_SUCCESS;
}

static int datax(int count, FILE *ifp, FILE *ofp, struct track_st *track)
{
  int status;

  for ( ; count; count--)
  {
    status = fgetc(ifp);
    if (status == EOF) {
      (void) printf("error: fgetc failure\n");
      return E_GET_CHAR;
    }
    (track->in)++;
    status = fputc(status, ofp);
    if (status == EOF) {
      (void) printf("error: fputc failure\n");
      return E_PUT_CHAR;
    }
    (track->out)++;
  }

  return STATUS_SUCCESS;
}

static int filec(FILE **ifp, FILE **ofp)
{
  int status;

  status = fclose(*ifp);
  if (status == EOF) {
    (void) printf("error: failed to close input file\n");
    return E_CLOSE_INPUT;
  }

  status = fclose(*ofp);
  if (status == EOF) {
    (void) printf("error: failed to close output file\n");
    return E_CLOSE_OUTPUT;
  }

  return STATUS_SUCCESS;
}
